package com.bergerkiller.bukkit.nolagg.sending;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;
import java.util.zip.Deflater;

import org.bukkit.entity.Player;
import org.getspout.spout.packet.listener.PacketListeners;
import org.getspout.spout.packet.standard.MCCraftPacket51MapChunk;

import com.bergerkiller.bukkit.nolagg.BufferedChunk;
import com.bergerkiller.bukkit.nolagg.NoLagg;
import com.bergerkiller.bukkit.nolagg.Util;

import net.minecraft.server.Chunk;
import net.minecraft.server.EntityPlayer;
import net.minecraft.server.Packet;
import net.minecraft.server.Packet51MapChunk;
import net.minecraft.server.TileEntity;

public class ChunkSendThread extends Thread {
	
	//thread creation
	private static ChunkSendThread[] threads;
	public static void init(int threadcount) {
		if (threads != null) {
			if (threads.length > threadcount) {
				//kill some threads
				ChunkSendThread[] newthreads = new ChunkSendThread[threadcount];
				for (int i = 0; i < threads.length; i++) {
					if (i < newthreads.length) {
						newthreads[i] = threads[i];
					} else {
						threads[i].interrupt();
					}
				}
				threads = newthreads;
			} else if (threads.length < threadcount) {
				ChunkSendThread[] newthreads = new ChunkSendThread[threadcount];
				System.arraycopy(threads, 0, newthreads, 0, threads.length);
				for (int i = threads.length; i < newthreads.length; i++) {
					newthreads[i] = new ChunkSendThread();
					newthreads[i].start();
				}
				threads = newthreads;
			}
			return;
		}
		threads = new ChunkSendThread[threadcount];
		for (; threadcount > 0; --threadcount) {
			threads[threadcount - 1] = new ChunkSendThread();
			threads[threadcount - 1].start();
		}
	}
	public static void deinit() {
		for (ChunkSendThread thread : threads) {
			thread.interrupt();
		}
		threads[0].finish();
		threads = null;
	}
	
	private final byte[] rawbuffer = new byte[81920];
	private final byte[] compbuffer = new byte[81920];
    private final Deflater deflater = new Deflater();
    
    private void put(Chunk chunk) {
		System.arraycopy(chunk.b, 0, this.rawbuffer, 0, 32768);
        System.arraycopy(chunk.g.a, 0, this.rawbuffer, 32768, 16384);
        System.arraycopy(chunk.i.a, 0, this.rawbuffer, 49152, 16384);
        System.arraycopy(chunk.h.a, 0, this.rawbuffer, 65536, 16384);
    }
    private void deflate(BufferedPacket51MapChunk packet) {
        this.deflater.reset();
        this.deflater.setLevel(6);
        this.deflater.setInput(packet.rawData);
        this.deflater.finish();
        packet.size = this.deflater.deflate(this.compbuffer);
        if (packet.size == 0) {
        	packet.size = this.deflater.deflate(this.compbuffer);
        }
        packet.prepareBuffer();
        packet.writeBuffer(this.compbuffer);
    }
    
    public static void queue(final EntityPlayer ep, final Chunk chunk, boolean lowpriority) {
    	QueuedPacket packet = new QueuedPacket(ep, chunk, lowpriority);
    	synchronized (lowPackets) {
    		if (lowpriority) {
    			lowPackets.offer(packet);
    		} else {
    			highPackets.offer(packet);
    		}
    	}
    }
    
    private static class QueuedPacket {
    	private static final List<Packet> tileEntityBuffer = new ArrayList<Packet>();
    	
    	@SuppressWarnings("unchecked")
		public QueuedPacket(final EntityPlayer ep, final Chunk chunk, final boolean lowPriority) {
    		this.destination = ep;
    		this.chunk = chunk;
    		Packet p;
    		for (TileEntity te : (Collection<TileEntity>)chunk.tileEntities.values()) {
    			p = te.k();
    			if (p != null) tileEntityBuffer.add(p);
    		}
    		this.tilepackets = tileEntityBuffer.toArray(new Packet[0]);
    		this.lowPriority = lowPriority;
    	}
    	private Packet[] tilepackets;
    	private final EntityPlayer destination;
    	private final Chunk chunk;
    	public final boolean lowPriority;

    	public boolean isNear() {
    		if (Math.abs(chunk.x - ((int) destination.locX >> 4)) > Util.view + 1) return false;
    		if (Math.abs(chunk.z - ((int) destination.locZ >> 4)) > Util.view + 1) return false;
    		return true;
    	}
    	
		public void send(final Packet51MapChunk packet) {
			synchronized (this.destination.netServerHandler) {
	    		this.destination.netServerHandler.sendPacket(packet);
	    		for (Packet p : this.tilepackets) {
	    			this.destination.netServerHandler.sendPacket(p);
	    		}
			}
    	}
		
		public BufferedPacket51MapChunk getPacket(final byte[] rawdata) {
			BufferedPacket51MapChunk packet = new BufferedPacket51MapChunk(this.chunk.x, this.chunk.z, rawdata);
			packet.lowPriority = this.lowPriority;
			return packet;
		}
    }
    
    
    private static final LinkedList<QueuedPacket> lowPackets = new LinkedList<QueuedPacket>();
    private static final LinkedList<QueuedPacket> highPackets = new LinkedList<QueuedPacket>();
    
    private void finish() {
    	synchronized (lowPackets) {
    		for (QueuedPacket packet : highPackets) {
    			this.handle(packet);
    		}
    		for (QueuedPacket packet : lowPackets) {
    			this.handle(packet);
    		}
    	}
    }
            
    private void handle(QueuedPacket packet) {
    	try {
    		if (packet.destination.dead) return;
    		if (packet.destination.netServerHandler.disconnected) return;
    		if (packet.destination.netServerHandler.networkManager == null) return;
    		if (!packet.isNear()) return;
    		Player player = packet.destination.getPlayer();
    		if (player == null) return;
    		if (packet.chunk instanceof BufferedChunk) {
    			if (((BufferedChunk) packet.chunk).isFree()) {
    				return;
    			}
    		}
    		this.put(packet.chunk);
    		//create packet
    		BufferedPacket51MapChunk mapchunk = packet.getPacket(this.rawbuffer);
    		//send chunk through spout if enabled
    		if (NoLagg.isSpoutEnabled) {
    			try {
    	    		MCCraftPacket51MapChunk mcpacket = new MCCraftPacket51MapChunk();
    	    		mcpacket.setPacket(mapchunk, 51);
    	    		PacketListeners.canSendUncompressedPacket(player, mcpacket);
    			} catch (Exception ex) {
    				ex.printStackTrace();
    			} catch (Throwable t) {
    				NoLagg.log(Level.SEVERE, "Spout support has been disabled: a fatal error occurred:");
    				t.printStackTrace();
    				NoLagg.isSpoutEnabled = false;
    			}
    		}
    		//==================================
    		//actually send this chunk
    	    this.deflate(mapchunk);
    	    packet.send(mapchunk);
    	} catch (Throwable t) {
    		NoLagg.log(Level.SEVERE, "Failed to send map chunk [" + packet.chunk.x + ", " + packet.chunk.z + "] to player " + packet.destination.name);
    		t.printStackTrace();
    	}
    }
    
    public static double getBusyPercentage(long timescale) {
    	double per = 0;
    	for (ChunkSendThread thread : threads) {
    		per += (double) thread.busyDuration / timescale;
    		thread.busyDuration = 0;
    	}
    	per /= threads.length;
    	return per;
    }
    
    public long busyDuration = 0;
    
    public final void run() {
    	QueuedPacket packet;
    	long oldtime;
    	while (!this.isInterrupted()) {
    		try {
    			oldtime = System.currentTimeMillis();
    			synchronized (lowPackets) {
    				packet = highPackets.poll();
    				if (packet == null) packet = lowPackets.poll();
    			}
    			if (packet == null) {
    				Thread.sleep(200);
    			} else {
    				this.handle(packet);
    				busyDuration += System.currentTimeMillis() - oldtime;
    			}
    		} catch (InterruptedException ex) {
    		} catch (Throwable t) {
    			t.printStackTrace();
    		}
    	}
    }

}
